/*
 *    MAKEIDX - (c) 1996 Mathijs Maassen (PE1NTP)
 *
 *  v1.0 - Oct 1993 (SM-Doze, in Pascal)
 *  v2.0 - Dec 1996 / Jan 1997 (Linux only)
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <ftw.h>
#include <dirent.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <getopt.h>
#include "fbb.h"

#define MAXDIRS  1024
#define MAXFILES 1024

struct fl {
	char name[13];
	size_t size;
	time_t time;
};

static char *verstr = "\n[ MAKEIDX/Linux v2.0 - (c) 1996 Mathijs Maassen PE1NTP @ PI8DAZ.#TWE.NLD.EU ]\n";
static char *dirs[MAXDIRS], *rootdir, *init_srv = "init.srv", *top;
static int rootdirlen, dircnt, labelcnt, no_drives;
static unsigned long total_bytes, total_files, total_dirs;
static YapLbl **labels;
static FILE *out = stdout;

/*
 *	Memory allocation with error checking/handling.
 */
static void *do_malloc(size_t size)
{
	void *ret;

	if ((ret = malloc(size)) == NULL) {
		fprintf(stderr, "Out of memory\n");
		exit(1);
	}

	return (ret);
}

/*
 *	qsort(3) compare function. Compare 2 **strings.
 */
static int sorttree(const void *s1, const void *s2)
{
	return (strcmp(*(char **) s2, *(char **) s1));
}

/*
 *	qsort(3) compare function. Compare 2 **fl fields.
 */
static int sortdir(const void *s1, const void *s2)
{
	return (strcmp((*(struct fl **) s2)->name, (*(struct fl **) s1)->name));
}

/*
 *	qsort(3) compare function. Compare 2 **YapLbl fields.
 */
static int sortlabels(const void *s1, const void *s2)
{
	return (strcmp((*(YapLbl **) s1)->filename, (*(YapLbl **) s2)->filename));
}

/*
 *	bsearch(3) compare function. Compare *str to **YapLbl field.
 */
static int searchlabel(const void *s1, const void *s2)
{
	return (strcmp((char *) s1, (*(YapLbl **) s2)->filename));
}

/*
 *	Map yapplbl.dat into memory. Create an array of pointers to the yapplbl
 *	records (labels[]) and sort them for use with bsearch(3).
 */
static void map_yapplbl(char *fname)
{
	struct stat sb;
	int fd, i;
	YapLbl *p;

	if ((fd = open(fname, O_RDONLY)) < 0) {
		perror("Warning: open() (yapplbl.dat)");
		return;
	}

	if (fstat(fd, &sb) < 0) {
		close(fd);
		perror("Warning: fstat() (yapplbl.dat)");
		return;
	}

	if ((p = mmap(0, sb.st_size, PROT_READ, MAP_SHARED, fd, 0)) == (YapLbl *) - 1) {
		close(fd);
		perror("Warning: mmap() (yapplbl.dat)");
		return;
	}

	if (!sb.st_size) {
		fprintf(stderr, "No labels in %s", fname);
		return;
	}

	labelcnt = sb.st_size / sizeof(YapLbl);
	labels = (YapLbl **) do_malloc(labelcnt * sizeof(YapLbl *));

	for (i = 0; i < labelcnt; i++)
		labels[i] = p++;

	qsort(labels, labelcnt, sizeof(YapLbl *), sortlabels);
}

/*
 *	Get a label from yapplbl.dat. The UNIX filename has to be converted first.
 */
static char *label_search(char drive, char *dirname, char *filename)
{
	int i;
	char fname[256];
	void *p;

	sprintf(fname, "%c:%s%s%s", drive, dirname, (*filename) ? "/" : "", filename);

	for (i = 0; fname[i]; i++) {
		if (fname[i] == '/')
			fname[i] = '\\';
		else
			fname[i] = toupper(fname[i]);
	}

	if ((p = bsearch(fname, labels, labelcnt, sizeof(YapLbl *), searchlabel)) == NULL)
		return ("");

	return ((*(YapLbl **) p)->label);
}

/*
 *	Called by ftw(3). Adds (non-hidden) directory to dirs[] array.
 */
static int add_dir(const char *file, struct stat *filestat, int flag)
{
	if (flag != FTW_D || file[rootdirlen + 1] == '.')
		return (0);

	dirs[dircnt] = (char *) do_malloc(strlen(&file[rootdirlen]) + 1);
	strcpy(dirs[dircnt], &file[rootdirlen]);
	if (++dircnt == MAXDIRS)
		return (-1);

	return (0);
}

/*
 *	Convert filename to DOS-filename (ie. uppercase and '\' instead of '/')
 */
static char *dosname(char *str)
{
	int i;

	for (i = 0; str[i]; i++) {
		if (str[i] == '/')
			str[i] = '\\';
		else
			str[i] = toupper(str[i]);
	}

	return (str);
}

/*
 *	Print the contents of a directory including labels, size and date.
 */
static void showdir(char drive, char *name)
{
	DIR *dir;
	struct dirent *entry;
	struct stat sb;
	struct fl *files[MAXFILES];
	char dirname[256], olddir[256], *p;
	int entries = 0;

	p = stpcpy(dirname, rootdir);
	strcpy(p, name);

	getcwd(olddir, sizeof(olddir));
	if (chdir(dirname) < 0) {
		perror(dirname);
		return;
	}

	if (no_drives)
		fprintf(out, "\n%s\\", dosname(name));
	else
		fprintf(out, "\n%c:%s\\", toupper(drive), dosname(name));

	p = label_search(drive, name, "");
	if (*p)
		fprintf(out, " [%s]", p);
	fputc('\n', out);

	if ((dir = opendir(".")) == NULL) {
		perror(dirname);
		chdir(olddir);
		return;
	}

	while ((entry = readdir(dir)) != NULL) {
		if (!strcmp(entry->d_name, "@@.lbl") || entry->d_name[0] == '.')
			continue;
		stat(entry->d_name, &sb);
		if (!S_ISREG(sb.st_mode))
			continue;
		files[entries] = (struct fl *) do_malloc(sizeof(struct fl));
		strncpy(files[entries]->name, entry->d_name, 12);
		files[entries]->name[12] = '\0';
		files[entries]->size = sb.st_size;
		files[entries]->time = sb.st_mtime;
		total_bytes += sb.st_size;
		total_files++;
		entries++;
	}

	qsort(files, entries, sizeof(struct fl *), sortdir);

	while (entries--) {
		struct tm *tm;
		char datestr[11];

		tm = localtime(&files[entries]->time);
		strftime(datestr, sizeof(datestr), "%d-%m-%Y", tm);
		fprintf(out, "    %-12s %7d %s %s\n", files[entries]->name, files[entries]->size, datestr, label_search(drive, name, files[entries]->name));
		free(files[entries]);
	}

	closedir(dir);
	chdir(olddir);
}

/*
 *	Get FBBDOS dirs (drives) and yapplbl filename from init.srv.
 */
void read_initsrv(char *path, char *fbbdos, char *yapplbl)
{
	FILE *fp;
	char line[256];
	int lines = 0;

	if ((fp = fopen(path, "r")) == NULL) {
		perror(path);
		exit(1);
	}

	while (fgets(line, sizeof(line), fp) != NULL) {
		if (line[0] == '#')
			continue;
		line[strlen(line) - 1] = '\0';
		switch (++lines) {
		case 5:{				/* system dir */
				char *p;

				p = stpcpy(yapplbl, line);
				strcpy(p, "yapplbl.dat");
				continue;
			}
		case 8:				/* fbbdos drives */
			strcpy(fbbdos, line);
			break;
		default:
			continue;
		}
		break;
	}

	fclose(fp);
}

/*
 *	Print usage info.
 */
void usage(char *prgname)
{
	printf("\nUsage: %s [OPTIONS]\n\n"
		"  -h, --help         This info\n"
		"  -i, --init file    Path to init.srv file (default=./init.srv)\n"
		"  -n, --nodrives     Don't display FBBDOS drive letters\n"
		"  -o, --output file  Output to file (default=stdout)\n"
		"  -t, --top file     Put file on top of index\n", prgname);
	exit(1);
}

int main(int argc, char **argv)
{
	int drive = 'a' - 1;
	char fbbdos[256], yapplbl[256];

	fputs(verstr, stderr);

	while (1) {
		struct option long_options[] =
		{
			{"help", no_argument, NULL, 'h'},
			{"init", required_argument, NULL, 'i'},
			{"nodrives", no_argument, NULL, 'n'},
			{"output", required_argument, NULL, 'o'},
			{"top", required_argument, NULL, 't'},
			{0, 0, NULL, 0}
		};

		switch (getopt_long(argc, argv, "hi:no:t:", long_options, NULL)) {
		case 'h':
			usage(argv[0]);
		case 'i':
			init_srv = optarg;
			continue;
		case 'n':
			no_drives = 1;
			continue;
		case 'o':
			if ((out = fopen(optarg, "w")) == NULL) {
				perror(optarg);
				exit(1);
			}
			continue;
		case 't':
			top = optarg;
			continue;
		default:
			optind = -1;
		case EOF:
			break;
		}
		if (optind < argc) {
			printf("Try `%s --help' for usage info.\n", argv[0]);
			exit(1);
		}
		break;
	}

	read_initsrv(init_srv, fbbdos, yapplbl);
	map_yapplbl(yapplbl);

	if (top && *top) {
		FILE *fp;
		int len;
		char buf[2048];

		if ((fp = fopen(top, "r")) == NULL) {
			perror(top);
			exit(1);
		}

		while ((len = fread(buf, 1, sizeof(buf), fp)) > 0)
			fwrite(buf, 1, len, out);

		fclose(fp);
	}

	rootdir = strtok(fbbdos, ",");
	do {
		drive++;
		if (rootdir[0] == '*')
			continue;

		dircnt = 0;
		rootdirlen = strlen(rootdir);
		if (rootdir[rootdirlen - 1] == '/') {
			rootdir[rootdirlen - 1] = '\0';
			rootdirlen--;
		}

		ftw(rootdir, add_dir, 240);
		qsort(dirs, dircnt, sizeof(char *), sorttree);

		total_dirs += dircnt;

		while (dircnt--) {
			showdir(drive, dirs[dircnt]);
			free(dirs[dircnt]);
		}

	} while ((rootdir = strtok(NULL, ",")) != NULL);

	fprintf(out, "\n%lu files, %lu bytes, %lu directories\n%s", total_files, total_bytes, total_dirs, verstr);

	return (0);
}
